import random

from .base import Aspect


class Punctuation(Aspect):
    """Randomly adds or removes specified punctuation marks.
    The implementation guarantees that punctuation marks won't be
    appended to the original ones or won't replace them after removal.

    With default settings all occurrences of the specified punctuation
    mark will be removed.

    - Example::

        Sentence, have a comma.

        Possible transformations:
        - Sentence have, a comma.
        - Sentence, have, a, comma.

        Impossible transformations:
        - Sentence,, have a comma.

    .. caution:: Uses random numbers, default seed is 42.
    """

    def __init__(self, char=',', add_percentage=0, remove_percentage=100, seed=42):
        """

        :param char: Punctuation mark that will be removed or
            added to sentences.

        :param add_percentage: Max percentage of white spaces in a sentence
            to be prepended with punctuation marks.

        :param remove_percentage: Max percentage of existing punctuation marks
            that will be removed.

        :param seed: Random seed.
        """

        if add_percentage >= 1:
            add_percentage /= 100.
        if remove_percentage >= 1:
            remove_percentage /= 100.

        if isinstance(char, str) and len(char) == 1:
            self._char = char
        else:
            self._char = ','
            print('Only single characters can be used as punctuation marks.')

        self._add_percentage = add_percentage
        self._remove_percentage = remove_percentage

        random.seed(seed)

    def __call__(self, sentence):

        modified_sentence = self._modify_punctuation(sentence)

        return modified_sentence

    def _modify_punctuation(self, sentence):
        punctuation_idx = []
        space_idx = []
        for i, char in enumerate(sentence):
            if char == self._char:
                punctuation_idx.append(i)
            elif char == ' ':
                try:
                    # Prevent doubling punctuation marks.
                    if i == punctuation_idx[-1] + 1:
                        continue
                except IndexError:
                    pass

                space_idx.append(i)

        random.shuffle(punctuation_idx)
        random.shuffle(space_idx)

        selected_punctuation = \
            sorted(punctuation_idx[:int(len(punctuation_idx) * self._remove_percentage)])
        selected_space = \
            sorted(space_idx[:int(len(space_idx) * self._add_percentage)])

        modified = ""
        for i, char in enumerate(sentence):
            if i in selected_punctuation:
                char = ''
            elif i in selected_space:
                char = self._char + ' '

            modified += char

        return modified



import math
import random

from .base import Aspect


class RemoveChar(Aspect):
    """Randomly removes characters from words.

    .. Note:: Note that you may specify white space as a character to be removed
              but it'll be processed differently.
    .. caution:: Uses random numbers, default seed is 42.
    """

    def __init__(self, char=None, words_percentage=50,
                 characters_percentage=10, seed=42):
        """

        :param words_percentage: Percentage of words in a
            sentence that should be transformed. If greater than 0,
            always at least single word will be transformed.

        :param characters_percentage: Percentage of characters in a
            word that should be transformed. If greater than 0
            always at least single character will be transformed.

        :param char: If specified only that character will be randomly removed.
             The specified character can also be a white space.

        :param seed: Random seed.
        """

        if words_percentage >= 1:
            words_percentage /= 100.

        if characters_percentage >= 1:
            characters_percentage /= 100.

        self._words_percentage = words_percentage
        self._characters_percentage = characters_percentage
        self._char = char

        random.seed(seed)

    def __call__(self, sentence):

        if self._char == ' ':
            return self._process_white_space(sentence)

        tokens = self._tokenize(sentence)
        modified_tokens = self._remove_characters(tokens)

        return self._detokenize(modified_tokens)

    def _remove_characters(self, tokens):

        # TODO: I think that punctuation marks etc. should
        #     be excluded.
        tokens_filtered = [i for i, token in enumerate(tokens)
                           if token.isalpha() or token.isdigit()]
        random.shuffle(tokens_filtered)

        selected_tokens =\
            sorted(tokens_filtered[:self._percentage_to_num(
                tokens_filtered, self._words_percentage)])

        modified = []
        for i, token in enumerate(tokens):
            if i in selected_tokens:

                if self._char:
                    token = token.replace(self._char, '',
                                          self._percentage_to_num(
                                              token,
                                              self._characters_percentage))
                else:
                    # TODO: Here's the change,
                    #  now all the characters are selected randomly.
                    selected_chars = [i for i, _ in enumerate(token)]
                    random.shuffle(selected_chars)

                    selected_chars =\
                        sorted(selected_chars[:self._percentage_to_num(
                            token, self._characters_percentage)])

                    modified_token = ""
                    for j, char in enumerate(token):
                        if j in selected_chars:
                            char = ''
                        modified_token += char

                    token = modified_token

            modified.append(token)

        return modified

    def _process_white_space(self, sentence):
        occurrences = int((sentence.count(' ') * self._words_percentage))

        return sentence.replace(' ', '', occurrences)

    @staticmethod
    def _percentage_to_num(array, percentage):
        if percentage == 0:
            return 0
        # Ensure that at least one item will be transformed.
        return max(1, int(len(array) * percentage))








import csv
import os

from .base import Dataset


class IMDB(Dataset):
    """The IMDB dataset containing movie reviews for a sentiment analysis.
    The dataset consists of 50 000 reviews of two classes, negative and positive.
    Each review is stored in a separate text file.
    For details see: http://ai.stanford.edu/~amaas/data/sentiment/
    """

    def load(self, path):
        """Loads a SNLI dataset.

        :param path: A path to single file, directory containing review files
                     or list of paths to such directories.

        :return: None
        """

        if type(path) is str and os.path.isdir(path):
            self._load_multiple_files(path)

        elif type(path) is list:
            for single_path in path:
                self._load_multiple_files(single_path)

        elif os.path.isfile(path):
            _, filename = os.path.split(path)
            entry = {'path': filename,
                     'content': self._read_file(path)}
            self._data.append(entry)

    def apply(self, aspect):
        """Modifies contents of the whole files in the IMDB dataset.
        """

        modified = []
        for entry in self._data:
            modified_sentence = aspect(entry['content'])
            entry['content'] = modified_sentence
            modified.append(entry)

        return modified

    def save(self, data, path):
        """Saves IMDB reviews to separate files
        with the original names.

        :param path: path to a top directory where files will be saved.

        :return: None
        """

        for entry in data:
            directory, filename = os.path.split(entry['path'])
            full_path = os.path.join(path, directory)
            if not os.path.exists(full_path) and directory != '':
                os.makedirs(full_path)
            with open(os.path.join(path, entry['path']), 'w') as f:
                f.write(entry['content'])

    def save_tsv(self, data, path):
        """Convenience function for saving IMDB reviews into a single TSV file.

        :param path: Path to a tab separated file.

        :return: None
        """

        with open(path, 'w', newline='') as tsvfile:
            writer = csv.writer(tsvfile, delimiter='\t')
            writer.writerow(['Sentiment', 'Content'])

            for entry in data:
                directory, _ = os.path.split(entry['path'])
                if directory == 'neg':
                    sentiment = 'neg'
                elif directory == 'pos':
                    sentiment = 'pos'
                else:
                    sentiment = 'unsup'

                writer.writerow([sentiment, entry['content']])

    @staticmethod
    def _read_file(path):

        with open(path, 'r') as f:
            content = f.read()

        return content

    def _load_multiple_files(self, path):

        filenames = os.listdir(path)
        for filename in filenames:
            full_path = os.path.join(path, filename)

            _, patent_dir = os.path.split(path)
            entry = {'path': os.path.join(patent_dir, filename),
                     'content': self._read_file(full_path)}
            self._data.append(entry)



from ..base import Aspect


class PigLatin(Aspect):
    """See wikipedia https://en.wikipedia.org/wiki/Pig_Latin
    """

    def __call__(self, sentence):
        return " ".join([self._transform_word(word)
                        for word in sentence.split()])

    @staticmethod
    def _transform_word(word):
        if len(word) == 0:
            raise ValueError("Can't transfer empty strings!")
        transformed_word = word[1:] + word[0] + 'ay'
        return transformed_word.lower()



from .base import Dataset


class SampleDataset(Dataset):
    """Simple list of sentences solely
    for testing and example purpose.
    """

    def __init__(self):
        self._data = None

    def load(self):
        self._data = [
            "Manning is a leader in applying Deep Learning "
            "to Natural Language Processing",
            "Manning has coauthored leading textbooks on statistical "
            "approaches to Natural Language Processing"
        ]

    def apply(self, aspect):
        return [aspect(sentence) for sentence in self._data]

    def save(self):
        pass



from abc import ABC
from functools import wraps


def file_exists_check(func):

    @wraps(func)
    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except FileNotFoundError:
            print("Check if the file path is correct")

    return func_wrapper


class Dataset(ABC):

    def __init__(self, *args, **kwargs):
        self._data = []

    @property
    def data(self):
        """Property of the Dataset class.

        :return: Internal object storing a loaded dataset.
        """
        return self._data

    def load(self, *args, **kwargs):
        """The method should handle loading
        and parsing of a specific dataset.
        """
        pass

    def apply(self, *args, **kwargs):
        """The method should iterate through texts
        in the dataset and apply a given aspect to them.
        """
        pass

    def save(self, *args, **kwargs):
        """The method should iterate through texts
        in the dataset and apply a given aspect to them.
        """
        pass



import random

from .base import Aspect


class Swap(Aspect):
    """Randomly swaps two characters within a word, excluding punctuations.
    It's possible that the same two characters will be swapped, so
    the word won't be changed, for example `letter` can become `letter`
    after swapping.

    .. caution:: Uses random numbers, default seed is 42.
    """

    def __init__(self, transform_percentage=100, seed=42):
        """

        :param transform_percentage: Maximum percentage of words in a
            sentence that should be transformed.

        :param seed: Random seed.
        """

        if transform_percentage > 1:
            transform_percentage /= 100.

        self._transform_percentage = transform_percentage

        random.seed(seed)

    def __call__(self, sentence):
        tokens = self._tokenize(sentence)

        modified_tokens = self._swap_characters(tokens)

        return self._detokenize(modified_tokens)

    def _swap_characters(self, tokens):

        # TODO: It's different than example in the paper
        #       where two letter words are also included.
        tokens_filtered = [i for i, token in enumerate(tokens)
                           if len(token) > 2]

        random.shuffle(tokens_filtered)
        selected_tokens =\
            sorted(tokens_filtered[:self._percentage_to_number(len(tokens))])

        modified = []
        for i, token in enumerate(tokens):
            if i in selected_tokens:
                idx_letter_first = random.randint(1, len(token) - 2)
                token = token[:idx_letter_first]\
                    + token[idx_letter_first + 1]\
                    + token[idx_letter_first]\
                    + token[idx_letter_first + 2:]

            modified.append(token)

        return modified

    def _percentage_to_number(self, num_tokens):
        return int(self._transform_percentage * num_tokens)




import json
import os
import random

from .base import Aspect


class QWERTY(Aspect):
    """Simulates errors made while writing on a QWERTY-type keyboard.
    Characters are swapped with their neighbors on the keyboard.

    .. caution:: Uses random numbers, default seed is 42.
    """

    def __init__(self, words_percentage=1, characters_percentage=10,
                 seed=42):
        """

        :param words_percentage: Percentage of words in a
            sentence that should be transformed. If greater than 0,
            always at least single word will be transformed.

        :param characters_percentage: Percentage of characters in a
            word that should be transformed. If greater than 0
            always at least single character will be transformed.

        :param seed: Random seed.
        """

        # TODO According to the original implementation
        #     it seem's that the variable should default to 1
        #     (when it was referring to absolute numbers)
        if words_percentage >= 1:
            words_percentage /= 100.

        if characters_percentage >= 1:
            characters_percentage /= 100.

        self._words_percentage = words_percentage
        self._characters_percentage = characters_percentage

        self._qwerty_mistakes = self._load_qwerty_mistakes()

        random.seed(seed)

    def __call__(self, sentence):
        tokens = self._tokenize(sentence)

        modified_tokens = self._transform_tokens(tokens)

        return self._detokenize(modified_tokens)

    def _transform_tokens(self, tokens):

        # TODO Differently to other aspects, here
        #    we don't filter out punctuations etc.
        tokens_idx = list(range(len(tokens)))
        random.shuffle(tokens_idx)

        selected_tokens =\
            sorted(tokens_idx[:self._percentage_to_num(
                tokens, self._words_percentage)])

        modified = []
        for i, token in enumerate(tokens):
            if i in selected_tokens:
                token = self._transform_token(token)
            modified.append(token)

        return modified

    def _transform_token(self, token):

        try:
            selected_chars = [i for i, _ in enumerate(token)]
            random.shuffle(selected_chars)

            selected_chars = \
                sorted(selected_chars[:self._percentage_to_num(
                    token, self._characters_percentage)])

            transformed_token = ""
            for i, char in enumerate(token):
                if i in selected_chars:
                    possible_mistakes =\
                        self._qwerty_mistakes[char.lower()]
                    mistake = random.choice(possible_mistakes)

                    if char.isupper():
                        char = mistake.upper()
                    else:
                        char = mistake

                transformed_token += char

        except KeyError:
            transformed_token = token

        return transformed_token

    @staticmethod
    def _load_qwerty_mistakes():

        current_dir = os.path.dirname(__file__)
        path = os.path.join(current_dir, 'auxiliary', 'qwerty.json')
        with open(path, 'r') as f:
            mistakes = json.load(f)

        return mistakes

    @staticmethod
    def _percentage_to_num(array, percentage):
        # Ensure that at least one item will be transformed.
        if percentage == 0:
            return 0
        return max(1, int(len(array) * percentage))






import json

from .base import Dataset, file_exists_check


class SQuAD(Dataset):
    """The SQuAD dataset.
    For details see: https://rajpurkar.github.io/SQuAD-explorer/
    """

    def __init__(self):
        self._data = None

    @file_exists_check
    def load(self, path):
        """Loads a SQuAD dataset.

        :param path: A path to a SQuAD data file in JSONL format.

        :return: None
        """
        with open(path, 'r') as f:
            self._data = json.load(f)

    def apply(self, aspect):
        """Modifies questions in the dataset
        leaving other data intact.
        """

        modified = self._data.copy()
        for entry in modified['data']:

            for paragraph in entry['paragraphs']:
                for qa in paragraph['qas']:
                    modified_sentence = aspect(qa['question'])
                    qa['question'] = modified_sentence

        return modified

    def save(self, data, path):
        """Saves data in the SQuAD format
        """

        with open(path, 'w') as f:
            json.dump(data, f)




from .articles import Articles
from .swap import Swap
from .qwerty import QWERTY
from .remove_char import RemoveChar
from .misspelling import Misspelling
from .digits2words import Digits2Words
from .sentiment_masking import SentimentMasking
from .punctuation import Punctuation



from io import open
import os
import random

from .base import Aspect


class SentimentMasking(Aspect):
    """This aspect reflects attempts made by Internet users
    to mask profanity or hate speech in online forums to evade moderation.
    We perform masking (replacing random, single character with for example an asterisk)
    of negative (or positive for completeness) words from Opinion Lexicon:
    http://www.cs.uic.edu/~liub/FBS/sentiment-analysis.html

    *All words that are listed will be transformed.*

    .. caution:: Uses random numbers, default seed is 42.
    """

    def __init__(self, char='*', use_positive=False, seed=42):
        """
        :param char: A character that will be used to mask words.

        :param use_positive: If True positive (instead of negative)
                             words will be masked.

        :param seed: Random seed.
        """

        if isinstance(char, str) and len(char) == 1:
            self._char = char
        else:
            self._char = '*'
            print('Only single characters should be used for masking.')

        filename = 'negative_words.txt'
        if use_positive:
            filename = 'positive_words.txt'
        self._sentiment_words = self._sentiment_words(filename)

        random.seed(seed)

    def __call__(self, sentence):
        tokens = self._tokenize(sentence)

        modified_tokens = self._transform_tokens(tokens)

        return self._detokenize(modified_tokens)

    def _transform_tokens(self, tokens):

        modified = []
        for token in tokens:
            if token.lower() in self._sentiment_words:
                ind = random.randint(0, len(token)-1)
                token = token[:ind] + self._char + token[ind+1:]
                # TODO: Appending a char instead of replacing is
                #   not implemented as it violates description
                #   from the paper.

            modified.append(token)

        return modified

    @staticmethod
    def _sentiment_words(filename):

        current_dir = os.path.dirname(__file__)
        path = os.path.join(current_dir, 'auxiliary', filename)

        with open(path, 'r', encoding='utf-8') as f:
            words = f.readlines()

        return [word.strip('\n') for word in words
                if word[0] != ';' and word != '\n']



import random

from .base import Aspect


class Articles(Aspect):
    """Randomly removes or swaps articles into wrong ones.

    .. caution:: Uses random numbers, default seed is 42.
    """

    def __init__(self, swap_probability=0.5, seed=42):
        """

        :param swap_probability: Probability of applying a transformation.
                                 Defaults to 0.5.

        :param seed: Random seed.
        """

        self._articles = ['a', 'an', 'the', '']

        if swap_probability > 1 or swap_probability < 0:
            swap_probability = 0.5
            print('swap_probability should be in range (0,1), '
                  'defaults to 0.5.')

        self._swap_probability = swap_probability
        random.seed(seed)

    def __call__(self, sentence):
        tokens = self._tokenize(sentence)
        swapped_tokens = self._swap_articles(tokens)

        return self._detokenize(swapped_tokens)

    def _swap_articles(self, tokens):

        modified = []
        for token in tokens:
            if token.lower() in self._articles \
               and random.random() < self._swap_probability:

                articles_copy = self._articles.copy()
                articles_copy.remove(token.lower())
                selected_article = random.choice(articles_copy)

                # TODO: Added handling the case when the article being
                #     modified is at the beginning of a sentence.
                if token[0].isupper():
                    selected_article = self._capitalize(selected_article)

                token = selected_article

            modified.append(token)

        return modified

    @staticmethod
    def _capitalize(token):
        try:
            return token[0].upper() + token[1:]
        except IndexError:
            return token





import json
import os
import random

from .base import Aspect


class Misspelling(Aspect):
    """Misspells words appearing in the Wikipedia list of
    **commonly misspelled English words** (default):
    https://en.wikipedia.org/wiki/Commonly_misspelled_English_words

    .. tip:: You can use **homophones** instead:
        https://en.wikipedia.org/wiki/Wikipedia:Lists_of_common_misspellings/Homophones

    If a word has more then one common misspelling, the replacement
    is selected randomly.

    *All words that have any misspellings listed will be replaced.*

    .. caution:: Uses random numbers, default seed is 42.
    """

    def __init__(self, use_homophones=False, seed=42):
        """
        :param use_homophones: If True list of homophones will be used
                               to replace words.

        :param seed: Random seed.
        """

        filename = 'misspellings.json'
        if use_homophones:
            filename = 'homophones.json'
        self._misspellings = self._load_misspellings(filename)

        random.seed(seed)

    def __call__(self, sentence):
        tokens = self._tokenize(sentence)

        modified_tokens = self._transform_tokens(tokens)

        return self._detokenize(modified_tokens)

    def _transform_tokens(self, tokens):

        modified = []
        for token in tokens:
            if token.lower() in self._misspellings:

                modified_token =\
                    random.choice(self._misspellings[token.lower()])
                if token[0].isupper():
                    modified_token = self._capitalize(modified_token)

                token = modified_token
            modified.append(token)

        return modified

    @staticmethod
    def _load_misspellings(filename):

        current_dir = os.path.dirname(__file__)
        path = os.path.join(current_dir, 'auxiliary', filename)
        with open(path, 'r') as f:
            mistakes = json.load(f)

        return mistakes

    @staticmethod
    def _capitalize(token):
        try:
            return token[0].upper() + token[1:]
        except IndexError:
            return token



import numpy as np

from .base import Dataset, file_exists_check


class CoNLL(Dataset):
    """The CoNLL-2003 shared task data for language-independent named
    entity recognition. For details see:
    https://www.clips.uantwerpen.be/conll2003/ner/
    """

    @file_exists_check
    def load(self, path):
        """Reads a CoNLL dataset file
        and loads into internal data structure in the following form:

        ::

            [{tokens: array(<tokens>)
              pos_tags: array(<pos_tags>),
              chunk_tags: array(<chunk_tags>),
              ner_tags: array(<ner_tags>},

              ...,

              ]

        :param path: A path to a file with CoNLL data

        :return: None

        """
        with open(path, "r") as f:
            for line in f.readlines():
                if line == "-DOCSTART- -X- O O\n":
                    continue

                elif line == "\n":
                    try:
                        processed = self._process_sample(sample)
                        self._data.append(processed)
                    except NameError:
                        pass
                    sample = ""

                else:
                    sample += line

            if sample != "":
                processed = self._process_sample(sample)
                self._data.append(processed)

    def apply(self, aspect, apply_to_ne=False):
        """

        :param aspect: transformation function

        :param apply_to_ne: if `False`, transformation won't be applied
                            to Named Entities. If `True`, transformation
                            will be applied only to Named Entities.

        :return: modified dataset in the following form:

        ::

            [{tokens: array(<tokens>)
              pos_tags: array(<pos_tags>),
              chunk_tags: array(<chunk_tags>),
              ner_tags: array(<ner_tags>},

              ...,

              ]

        """

        modified = []
        for entry in self._data:
            tags = entry['ner_tags']

            if apply_to_ne is False:
                non_ner = np.where(tags == 'O')[0]
            else:
                non_ner = np.where(tags != 'O')[0]

            if len(non_ner) == 0:
                modified.append(entry)
            else:
                sentence = " ".join(entry['tokens'][non_ner])
                modified_sentence = aspect(sentence)

                for idx, token in zip(non_ner, modified_sentence.split()):
                    entry['tokens'][idx] = token

                modified.append(entry)

        return modified

    def save(self, data, path):
        """Saves data in the CoNLL format

        :param data: list of dictionaries in the following form:

        ::

            [{tokens: array(<tokens>)
              pos_tags: array(<pos_tags>),
              chunk_tags: array(<chunk_tags>),
              ner_tags: array(<ner_tags>},

              ...,

              ]

        :param path: Path to save the file. If the file exists,
                     it will be overwritten.

        :return: None
        """

        with open(path, 'w+') as f:

            lines = list()
            lines.append('-DOCSTART- -X- O O\n')

            try:
                for entry in data:
                    line = ""
                    for token, pos_tag, chunk_tag, ner_tag \
                        in zip(entry['tokens'], entry['pos_tags'],
                               entry['chunk_tags'], entry['ner_tags']):
                        line += " ".join([token, pos_tag, chunk_tag, ner_tag]) + '\n'
                    lines.append(line)

                # Remove a trailing newline
                lines[-1] = lines[-1][:-1]
                f.write('\n'.join(lines))
            except KeyError:
                print("The data you're trying to save is corrupted or "
                      "isn't formatted correctly.")

    @staticmethod
    def _process_sample(sample):

        data = sample.split('\n')

        tokens = []
        pos_tags = []
        chunk_tags = []
        ner_tags = []

        for entry in data:
            if entry == '':
                continue

            info = entry.split()
            tokens.append(info[0])
            pos_tags.append(info[1])
            chunk_tags.append(info[2])
            ner_tags.append(info[3])

        processed = dict()
        processed['tokens'] = np.asarray(tokens)
        processed['pos_tags'] = np.asarray(pos_tags)
        processed['chunk_tags'] = np.asarray(chunk_tags)
        processed['ner_tags'] = np.asarray(ner_tags)

        return processed



from abc import ABC
import re


class Aspect(ABC):
    """Base, abstract class. All the aspects
    must implement the __call__ method.
    """

    def __call__(self, sentence):
        """We want to directly call
        objects of the Aspect class for easy chaining.
        This function will be applied to sentences.
        """
        pass

    @staticmethod
    def _tokenize(sentence):
        """Split text into tokens including punctuation
        and special characters.

        :param sentence: A sentences as a single string.

        :return: List of tokens.
        """
        return re.findall(r"[\w\'\-()]+|[^\s\w]", sentence)

    @staticmethod
    def _detokenize(tokens):
        """Join tokens into tokens including punctuation
        and special characters.

        :param tokens: List of tokens.

        :return: A sentence as a single string.
        """

        sentence = " ".join(tokens)

        # remove space between words and punctuation
        sentence = re.sub(r'\s([?.!,:;\'\"\\](?:\s|$))', r'\1', sentence)

        # remove double spaces
        return " ".join(sentence.split())





import json

from .base import Dataset, file_exists_check


class SNLI(Dataset):
    """The SNLI dataset supporting the task of natural language inference.
    For details see: https://nlp.stanford.edu/projects/snli/
    """

    @file_exists_check
    def load(self, path):
        """Loads a SNLI dataset.

        :param path: A path to a SNLI data file in JSONL format.

        :return: None
        """
        with open(path, 'r') as f:
            for line in f.readlines():
                self._data.append(json.loads(line))

    def apply(self, aspect):
        """Modifies premises (sentence1) in the dataset
        leaving other data intact.
        """

        # TODO: modify parsed sentences as well.

        modified = []
        for entry in self._data:
            try:

                modified_sentence = aspect(entry['sentence1'])
                entry['sentence1'] = modified_sentence
                modified.append(entry)

            except KeyError:
                print("Encountered a KeyError. The entry won't be modified")
                modified.append(entry)

        return modified

    def save(self, data, path):
        """Saves data in the SNLI format
        """
        json_strings = [json.dumps(entry) for entry in data]

        with open(path, 'w') as f:
            f.write("\n".join(json_strings))




import os
from setuptools import setup, find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="wild-nlp",
    version="0.0.4",
    author="Dominika Basaj, Barbara Rychalska, Alicja Gosiewska, Adam Słucki",
    author_email="adam.slucki@gmail.com",
    description=("Text aspects for nlp models"),
    license="BSD",
    url="https://github.com/MI2DataLab/WildNLP",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Text Processing",
        "License :: OSI Approved :: BSD License",
    ],
    install_requires=[
        'numpy',
        'num2words'
    ]
)



import functools


def compose(*functions):
    """Chains multiple aspects into a single function.

    :param functions: an arbitrary object(s) of the Callable instance.

    :return: chained function
    """
    return functools.reduce(lambda f, g: lambda x: g(f(x)),
                            functions, lambda x: x)



import re

from num2words import num2words

from .base import Aspect


class Digits2Words(Aspect):
    """Converts numbers into words.
    Handles floating numbers as well.

    *All numbers will be converted*
    """

    def __call__(self, sentence):
        tokens = self._tokenize(sentence)

        modified_tokens = self._transform_tokens(tokens)

        return self._detokenize(modified_tokens)

    @staticmethod
    def _transform_tokens(tokens):

        modified = []
        for token in tokens:

            append_dot = False
            token_modified = token

            # Handle case when a number is at the end of a sentence.
            if token_modified[-1] == '.':
                token_modified = token_modified[:-1]
                append_dot = True

            # Handle US format like 123.123,50
            if all(char in token for char in ['.', ',']):
                token_modified = token.replace('.', '').replace(',', '.')

            try:
                number = float(token_modified)
                token_modified = num2words(number)

                # TODO: It would be nice to uppercase
                #    the token if it's the first word in a sentence.
                #    Not so easy.

            except ValueError:
                pass

            modified.append(token_modified)

            if append_dot:
                modified.append('.')

        return modified

    @staticmethod
    def _tokenize(sentence):
        """Split text into tokens including punctuation
        and special characters.

        :param sentence: A sentences as a single string.

        :return: List of tokens.
        """
        return re.findall(r"[\w\'-.,]+|[^\s\w]", sentence)





from ..base import Aspect


class Reverser(Aspect):
    """Reverse all characters in words.
    """

    def __call__(self, sentence):
        return " ".join([self._transform_word(word)
                         if word != '' else ''
                         for word in sentence.split(' ')])

    @staticmethod
    def _transform_word(word):
        if len(word) == 0:
            raise ValueError("Can't transfer empty strings!")
        return word[::-1]
